/* IluJava_IluClassRep.c */
/* Chris Jacobi, April 9, 1997 5:05 pm PDT */

/*
 * Copyright (c) 1996, 1997 Xerox Corporation.  All Rights Reserved.  
 * Unlimited use, reproduction, and distribution of this software is
 * permitted.  Any copy of this software must include both the above
 * copyright notice of Xerox Corporation and this paragraph.  Any
 * distribution of this software must comply with all applicable United
 * States export control laws.  This software is made available AS IS,
 * and XEROX CORPORATION DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED,
 * INCLUDING WITHOUT LIMITATION THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE, AND NOTWITHSTANDING ANY OTHER
 * PROVISION CONTAINED HEREIN, ANY LIABILITY FOR DAMAGES RESULTING FROM
 * THE SOFTWARE OR ITS USE IS EXPRESSLY DISCLAIMED, WHETHER ARISING IN
 * CONTRACT, TORT (INCLUDING NEGLIGENCE) OR STRICT LIABILITY, EVEN IF
 * XEROX CORPORATION IS ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
 */
 
/* $Id: IluJava_IluClassRep.c,v 1.19 1997/04/10 01:34:09 jacobi Exp $ */
 
/*
 * C side of IluClassRep.java
 * also accessing IluMethodRep.java and IluMethodArgRep.java
 */


#include <iluxport.h>
#include <iluerror.h>

#include "IluJava_JStubs.h"
#include "xerox_ilu_IluMethodArgRep.h"
#include "xerox_ilu_IluExceptionRep.h"

#include "IluJava_JTypes.h"
#include "IluJava.h"



#define JToCStringKeep(obj, field)	allocCString( unhand(obj) -> field )
    /* Access a String field of an object, convert it to an
     * ilu_string and permanently allocate the ilu_string.
     * (javaString.h)allocCString does unicode to char conversion
     */
     
#define JToCStringGC(obj, field)	makeCString( unhand(obj) -> field )
    /* Access a String field of an object, convert it to an
     * ilu_string and make it subject to java garbage collection.
     * (javaString.h)allocCString does unicode to char conversion
     */
     
#define JToBoolean(obj, field)	(( unhand(obj) -> field ) != 0)
    /* Access a field of an object as a boolean */

/* Compare to /tilde/jacobi/ilu/runtime/python/iluclobject.c */


static void
freeSuperClasses(ilu_string *cSuperClassArray, int sccnt)
    /* specifically used in nativeFinishClass */
{
    if (cSuperClassArray) {
        int sx;
        for (sx = 0; sx < sccnt; sx = sx+1) {
            /* use java's free because strings were allocated by java */
            java_free(cSuperClassArray[sx]);
        }
        /* ilu's free as this was allocated with ilu */
        ilu_free(cSuperClassArray);
    }
}


EXPORTJAVA void 
xerox_ilu_IluClassRep_nativeFinishClass(
		struct Hxerox_ilu_IluClassRep * jjIluClass,
		Jint isRootClass
		)
{
    int mx, sx, ex, ax; /*loop indices*/
    ilu_Error err = ILU_INIT_NO_ERR;
    ilu_Error err2 = ILU_INIT_NO_ERR;
    ilu_Class cIluClass;
    ilu_string *cSuperClassArray = 0;
    ilu_string cIluClassName;
    ilu_string cUID;
    /*struct Hxerox_ilu_IluClassRep * jjSuper;*/
    struct HArrayOfObject * jjSuper;
    struct HArrayOfObject * jjMethods;
    struct HArrayOfObject * jjArgs; 
    int sccnt = 0;
    ilu_string cSingleton;
    int argCnt;
    
    if (isRootClass) {
        cIluClass = (ilu_Class) ilu_rootClass;
        Access(jjIluClass, yIluClass, ilu_Class) = cIluClass;
        _ilujava_registerJavaClass(cIluClass, jjIluClass);
        return;
    }
    
    cIluClassName = JToCStringKeep(jjIluClass, jjIluClassName);
    cUID = JToCStringKeep(jjIluClass, jjuid);
    jjMethods = unhand(jjIluClass)->jjMethods;
    
    /* superclass array */
    jjSuper = unhand(jjIluClass)->jjIluSuperClasses;
    if (jjSuper) sccnt = obj_length(jjSuper);
    if (_ilujava_definitionsFlag>0) {
        ilu_DebugPrintf("$ DefineClass %s sccnt: %d\n", cIluClassName, sccnt);
    }
    if (sccnt) {
        cSuperClassArray = ilu_must_malloc(sccnt * sizeof(ilu_string));
        for (sx = 0; sx < sccnt; sx++) {
            ilu_string sClassId; 
            struct Hxerox_ilu_IluClassRep* jjThisSuper;
            jjThisSuper = (struct Hxerox_ilu_IluClassRep*) 
                unhand(jjSuper)->body[sx];
            sClassId = /*javaString.h*/ allocCString(
                (Hjava_lang_String *) unhand(jjThisSuper)->jjuid
                ); /* C strings in C heap ! */
            cSuperClassArray[sx] = sClassId;
            if (_ilujava_definitionsFlag>0) {
               ilu_DebugPrintf("$ Superclass id [%s] \n", sClassId);
            }
        }
    }

    /* Start the kernel registrations */
    ilu_EnterMutex(ilu_otmu, &err);
    if (ILU_ERRNOK(err)) {
        freeSuperClasses(cSuperClassArray, sccnt);
        _ilujava_IluErrorToException(&err, completedNo "reallyFinish0");
        return;
    }
    
    /* The object class */
    cSingleton = ( unhand(jjIluClass)->jjsingleton==0 ? 
    	0 : JToCStringKeep(jjIluClass, jjsingleton)
    	);
    cIluClass = ilu_DefineObjectType(
        cIluClassName, 	/*cl_name*/
	JToCStringKeep(jjIluClass, jjbrand),	/*cl_brand*/
	cUID,   	/*cl_unique_id*/
    	cSingleton,
    	JToBoolean(jjIluClass, optional),
    	JToBoolean(jjIluClass, collectible),
    	JToCStringKeep(jjIluClass, docString),
    	(ilu_cardinal) unhand(jjIluClass)->method_count, 
    	(ilu_cardinal) sccnt, 
	cSuperClassArray, 
	&err
    	);
    Access(jjIluClass, yIluClass, ilu_Class) = cIluClass;
    if (ILU_ERRNOK(err)) {
        freeSuperClasses(cSuperClassArray, sccnt);
        _ilujava_IluErrorToException(&err, "reallyFinish1");
        return;
    }

    /* loop to register methods */
    for (mx = 0; mx < unhand(jjIluClass)->method_count; mx = mx+1) 
    {
        int exceptCnt = 0;
        ilu_Exception *cExceptionVec = 0;
        ilu_Method cIluMethod;
        struct Hxerox_ilu_IluMethodRep* jjMethRep
        	= (struct Hxerox_ilu_IluMethodRep*) unhand(jjMethods)->body[mx];
        struct HArrayOfObject * jjExceptions 
        	= unhand(jjMethRep)->jjExceptions;
        Hjava_lang_String * jjRetValUid 
        	= unhand(jjMethRep)->jjRetValUid;
        /* build exception array for this method */
        if (jjExceptions != 0) exceptCnt = obj_length(jjExceptions);
        if (exceptCnt) {
            cExceptionVec = 
        	ilu_must_malloc(exceptCnt * sizeof(ilu_Exception));
            for (ex = 0; ex < exceptCnt; ex = ex+1) 
            {
                struct Hxerox_ilu_IluExceptionRep* jExc;
                jExc = (struct Hxerox_ilu_IluExceptionRep*) 
                    unhand(jjExceptions)->body[ex];
                cExceptionVec[ex] = Access(
                	jExc,
                	yIluException, /*field*/
                	ilu_Exception  /*type*/
                	);
                	/* check against nil in java */
            }
        }
        /* register method */
        argCnt = (int) unhand(jjMethRep)->argCnt;
        ILU_CLER(err);
        cIluMethod = ilu_DefineMethod(
        	cIluClass,
        	(ilu_cardinal) mx,  /* i */
        	JToCStringGC(jjMethRep, jjIluMethodName), 
        	unhand(jjMethRep)->id,
                JToBoolean(jjMethRep, cacheable),
		JToBoolean(jjMethRep, asynchronous),
		(ilu_cardinal) exceptCnt, 
		cExceptionVec, 
		argCnt, 
		(jjRetValUid==0 ? 0 : makeCString(jjRetValUid)), 
		&err
        	);
        if (ILU_ERRNOK(err)) {
            ilu_free(cExceptionVec);
            freeSuperClasses(cSuperClassArray, sccnt);
            _ilujava_IluErrorToException(&err, "reallyFinish2");
            return;
        }
        Access(jjMethRep, yIluMethod, ilu_Method) = cIluMethod;
        
        /* Loop for defining the arguments of the method */
        jjArgs = unhand(jjMethRep)->jjArgs;
        for (ax = 0; ax < argCnt; ax = ax+1) 
        {
            ilu_boolean b;
            struct Hxerox_ilu_IluMethodArgRep* jjMethArg;
            jjMethArg = (struct Hxerox_ilu_IluMethodArgRep*) unhand(jjArgs)->body[ax];
            ILU_CLER(err);
            b = ilu_DefineMethodArg(
                cIluMethod,
                (ilu_cardinal) ax,
                JToCStringKeep(jjMethArg, argName),
                unhand(jjMethArg)->sibling,
                (ilu_ArgDirection) unhand(jjMethArg)->argDirection,
                JToCStringKeep(jjMethArg, typeUid),
                &err
                );
            if (ILU_ERRNOK(err)) {
                ilu_free(cExceptionVec);
                freeSuperClasses(cSuperClassArray, sccnt);
                _ilujava_IluErrorToException(&err, "reallyFinish3");
                /* leaking memory used for string arguments 
                 * of ilu_DefineMethodArg; I see no alternative.
                 */
                return;
            }
        }
        
        
        /* 
         * TERRIBLE LOOPHOLE 
         * Assigns a java object into a field declared as procedure 
         */
        ilu_SetMethodStubProc(cIluMethod, (ilu_StubProc) jjMethRep);
        ilu_free(cExceptionVec);
    }
    
    _ilujava_registerJavaClass(cIluClass, jjIluClass); /*IluJava.h*/
    
    ilu_ObjectTypeDefined(cIluClass, &err2);
    if (ILU_ERRNOK(err2)) {
        _ilujava_IluErrorToException(&err2, "reallyFinish98");
        goto return_exitMutexAndFreeSuperClasses;
    }

#ifdef IIOP_PROTOCOL
    {
        /*L1 < otmu required and no lock is held ... */
        ilu_boolean new;
        ilu_string cIfName;
        ilu_string cIfBrand = 0;
        cIfName = JToCStringKeep(jjIluClass, jjIfName);
        if (unhand(jjIluClass)->jjIfBrand) {
            cIfBrand = JToCStringKeep(jjIluClass, jjIfBrand);
        }
        ILU_CLER(err);
        ilu_RegisterObjectType(cIluClassName, 
            cIfName, cIfBrand, cUID, cIluClass, &new, &err
            );
        java_free(cIfName);
        if (cIfBrand) java_free(cIfBrand);
    }
#endif /* def IIOP_PROTOCOL */
    
    /*goto exitMutexAndFreeSuperClasses;*/
    
    return_exitMutexAndFreeSuperClasses:
        ILU_CLER(err);
        ilu_ExitMutex(ilu_otmu, ilu_FALSE, &err);
        if (ILU_ERRNOK(err)) {
            /*raise an java exception only if no other exception is pending */ 
            struct execenv *ee = EE();
            if (! ee->exceptionKind) {
                _ilujava_IluErrorToException(&err, "reallyFinish99");
            }
        }
        /*fall through*/
    return_freeSuperClasses:
        freeSuperClasses(cSuperClassArray, sccnt);
        /*fall through*/
    return_simple:
        return;
    
} /* nativeFinishClass */


EXPORTJAVA void 
xerox_ilu_IluExceptionRep_registerException(
		struct Hxerox_ilu_IluExceptionRep * jjIluExceptionRep)
{
    ilu_Exception cIluException;
    ilu_Error	err = ILU_INIT_NO_ERR;
    /* 
     * (javaString.h)makeCString converts unicode to char's 
     * and returns a garbage collectable C string
     * (ilu_DefineException says: caller owns data) 
     */
    char* i = makeCString(unhand(jjIluExceptionRep)->i); /*javaString.h*/
    char* e = makeCString(unhand(jjIluExceptionRep)->e); 
    char* valUid = 0; 
    if (unhand(jjIluExceptionRep)->jtuuid) {
        valUid = makeCString(unhand(jjIluExceptionRep)->jtuuid);
    }
    cIluException = ilu_DefineException(i, e, valUid, &err);
    ILU_MUST_BE_SUCCESS(err);
    Access(jjIluExceptionRep, yIluException, ilu_Exception) = cIluException;
    /* make sure pointers are around and not garbage collected to early */
}


/* end */
