/* IluJava_IluConnOps.c */
/* Chris Jacobi, April 14, 1997 5:00 pm PDT */

/*
 * Copyright (c) 1996 Xerox Corporation.  All Rights Reserved.  
 * Unlimited use, reproduction, and distribution of this software is
 * permitted.  Any copy of this software must include both the above
 * copyright notice of Xerox Corporation and this paragraph.  Any
 * distribution of this software must comply with all applicable United
 * States export control laws.  This software is made available AS IS,
 * and XEROX CORPORATION DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED,
 * INCLUDING WITHOUT LIMITATION THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE, AND NOTWITHSTANDING ANY OTHER
 * PROVISION CONTAINED HEREIN, ANY LIABILITY FOR DAMAGES RESULTING FROM
 * THE SOFTWARE OR ITS USE IS EXPRESSLY DISCLAIMED, WHETHER ARISING IN
 * CONTRACT, TORT (INCLUDING NEGLIGENCE) OR STRICT LIABILITY, EVEN IF
 * XEROX CORPORATION IS ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
 */
 
/* $Id: IluJava_IluConnOps.c,v 1.16 1997/04/15 22:43:58 jacobi Exp $ */
 
/* 
 * C side for IluSurrogateConnection.java, IluServerConnection.java
 */
  
#include <StubPreamble.h>
#include <interpreter.h>

#include <iluxport.h>
#include <iluerror.h>

#include "xerox_ilu_IluSurrogateConnection.h"
#include "xerox_ilu_IluServerConnection.h"
#include "xerox_ilu_IluPort.h"
#include "xerox_ilu_IluCall.h"
#include "xerox_ilu_IluClassRep.h"
#include "xerox_ilu_IluMethodRep.h"

#include "IluJava.h"
#include "IluJava_JTypes.h"

#define JIluServerConnection struct Hxerox_ilu_IluServerConnection*
#define JIluSurrogateConnection struct Hxerox_ilu_IluSurrogateConnection*
#define JIluPort struct Hxerox_ilu_IluPort*

#define TRACE0 _ilujava_pctFlag > 0
#define TRACE1 _ilujava_pctFlag > 1

static JIluSurrogateConnection jProtoCFInstance = 0;
    /* Some good instance from initialization because we failed 
     * calling static methods from native code.
     */


#define PORTWATCHER_LOOP 0
#define PORTWATCHER_FORK 1
#define PORTWATCHER_CLOSE 2
#define PORTWATCHER_DONE 3

EXPORTJAVA Jint 
xerox_ilu_IluPort_nativeWatchPort(
    	JIluPort jjPort
    	)
{
    ilu_boolean     ret;
    ilu_boolean     closed = 0;
    ILU_ERRS((IoErrs, bad_locks, broken_locks, no_resources, interrupted)) err;
    ilu_Connection  conn;
    ilu_Port cPort = Access(jjPort, yPort, ilu_Port);
    if (TRACE0) {
        ilu_DebugPrintf("$ IluConnOps_nativeWatchPort\n");
        ilu_DebugPrintf("   jjPort: %x, cPort: %x \n", (int)jjPort, (int)cPort);
    }
    ILU_CLER(err);
    ret = ilu_WaitForPortConnectionRequest(cPort, &err);
    if (ret==0) {
        Access(jjPort, yConn, ilu_Connection) = 0;
        ILU_ERR_SWITCH(err) {
            ILU_SUCCESS_CASE
	  	/* port is closed */
	  	if (TRACE0) {
                    ilu_DebugPrintf("$ .._nativeWatchPort port closed\n");
                }
                return PORTWATCHER_DONE;
            ILU_ERR_CASE(interrupted, v)
	  	/* interrupted; retry */
	  	if (TRACE0) {
                    ilu_DebugPrintf("$ .._nativeWatchPort interrupted\n");
                }
                return PORTWATCHER_LOOP;
            ILU_ERR_ELSE
	  	/* error; give up */
                if (TRACE0) {
                    ilu_DebugPrintf("$ .._nativeWatchPort ERR1\n");
                }
                return PORTWATCHER_DONE;
        } ILU_ERR_ENDSWITCH;
    }
    ILU_CLER(err);
    conn = ilu_HandleNewConnection(cPort, &closed, &err);
    if (closed) {
        if (TRACE0) {
            ilu_DebugPrintf("$ .._nativeWatchPort: closed \n");
        }
        Access(jjPort, yConn, ilu_Connection) = 0;
        return PORTWATCHER_DONE;
    }
    if (conn==0) {
        ILU_ERR_SWITCH(err) {
            ILU_SUCCESS_CASE
	  	/* spurious wakeup; retry */
	  	if (TRACE0) {
                    ilu_DebugPrintf("$ .._nativeWatchPort sprs\n");
                }
                Access(jjPort, yConn, ilu_Connection) = 0;
                return PORTWATCHER_LOOP;
            ILU_ERR_ELSE
	  	/* error; give up */
                if (TRACE0) {
                    ilu_DebugPrintf("$ .._nativeWatchPort ERR2\n");
                }
                if (Access(jjPort, yConn, int) == -1) {
                    /*prevent infinite loop*/
                    return PORTWATCHER_DONE;
                }
                Access(jjPort, yConn, int) = -1;
                return PORTWATCHER_CLOSE;
        } ILU_ERR_ENDSWITCH;
    } else {
        if (TRACE0) {
            ilu_DebugPrintf("$ .._nativeWatchPort: scss:<%p>\n",  conn);
        }
        Access(jjPort, yConn, ilu_Connection) = conn;
        /* Requirement: ilu_DoneServingConnection must called eventually.
         * Loop reading requests will do so.
         */
        return PORTWATCHER_FORK;
    }
    /* notreached */
} /* nativeWatchPort */


EXPORTJAVA Jboolean 
xerox_ilu_IluSurrogateConnection_nativeOtherNewConnection(
	JIluSurrogateConnection jjSurrConn
	)
{
    ilu_Connection cConn; 
    ILU_ERRS((IoErrs, internal)) err = ILU_INIT_NO_ERR;
    if (TRACE0) {
        ilu_DebugPrintf("$ IluConnOps_nativeOtherNewConnection");
    }
    cConn = ilu_OtherNewConnection(&err);
    if (ILU_ERRNOK(err)) {
        _ilujava_IluErrorToException(&err, "OtherNewConnection");
    }
    Access(jjSurrConn, ySurrConn, ilu_Connection) = cConn;
    return (cConn != 0);
} /* nativeOtherNewConnection */


EXPORTJAVA void 
xerox_ilu_IluSurrogateConnection_nativeNewConnGetterForked(
	JIluSurrogateConnection jjSurrConn
	)
{
    ILU_ERRS((IoErrs, internal)) err = ILU_INIT_NO_ERR;
    if (TRACE0) {
        ilu_DebugPrintf("$ IluConnOps_nativeNewConnGetterForked");
    }
    jProtoCFInstance = jjSurrConn;
    ilu_NewConnectionGetterForked(&err);
    ILU_MUST_BE_SUCCESS(err);
} /* nativeNewConnGetterForked */


EXPORTJAVA void 
xerox_ilu_IluSurrogateConnection_nativeOutgoingConnectionThreadProc(
	JIluSurrogateConnection jjSurrConn
	)
/* Called through forked Java when ilu_StartCall did require  
 * forking a connection reader handling one surrogate connection.
 * Uses "this.ySurrCon"
 */ 
{
    ILU_ERRS((IoErrs, internal)) err = ILU_INIT_NO_ERR;
    ilu_Connection cConn;
    cConn = Access(jjSurrConn, ySurrConn, ilu_Connection);
    if (TRACE0) {
        ilu_DebugPrintf("$ IluConnOps_NOCTP enter jjSurrConn %x cConn %x\n", 
            (int) jjSurrConn, cConn);
    }
    ilu_OutgoingConnectionThreadProc(cConn, &err);
    if (TRACE0) {
        ilu_DebugPrintf("$ IluConnOps_NOCTP exit jjSurrConn %x cConn %x\n", 
            (int) jjSurrConn, cConn);
    }
    ILU_MUST_BE_SUCCESS(err);
} /* nativeOutgoingConnectionThreadProc */


EXPORTJAVA Jboolean
xerox_ilu_IluServerConnection_receiveRequest(
    	JIluServerConnection jjIluServerConnection,
    	JIluCall jjcall
    	)
/* servicing requests for a true servers connection */
{
    ilu_boolean     initted;
    ilu_Call cCall = Access(jjcall, yCall, ilu_Call);
    ilu_Class       cIluClass;
    ilu_Method      cIluMethod;
    ilu_RcvReqStat  stat;
    ilu_cardinal    SN;
    ILU_ERRS((bad_locks, IoErrs, bad_param, broken_locks, internal)) lerr;
    ilu_boolean     closed = ilu_FALSE;
    ilu_Connection cConn = Access(jjIluServerConnection, yConn, ilu_Connection);
    if (TRACE0) {
        ilu_DebugPrintf("$ IluConnOps_receiveRequest cConn %p \n", cConn);
    }
    if (unhand(jjcall)->needFinish) {
        BUGPRINT("$ error with call");
        return 0; /* ? abandon connection */
    }
    unhand(jjcall)->jjMethodRep = 0;
    if (! ilu_BlockingWaitForInputOnConnection(cConn, ILU_NIL)) {
        /*connection should be abandoned*/
        if (TRACE0) {
            ilu_DebugPrintf("$ IluConnOps_receiveRequest abandon conn\n");
        }
        ilu_DoneServingConnection(cConn, &lerr);
        return 0;
    }
    if (TRACE1) {
        ilu_DebugPrintf("$ IluConnOps_receiveRequest_1 \n");
    }
    stat = ilu_ReceiveRequest(cCall, &initted, cConn, &cIluClass, &cIluMethod,
			    &SN, &lerr);
    if (stat == ilu_RcvReqStat_request) {
        /*initted is true*/
        if (TRACE1) {
            ilu_DebugPrintf("$ IluConnOps_receiveRequest request \n");
        }
        unhand(jjcall)->jMustFork = (Jboolean) ilu_ThreadPerRequest(cConn);
        unhand(jjcall)->jjClassRep = (JIluClassRep)
        	 _ilujava_findJavaClass(cIluClass);
        unhand(jjcall)->jjMethodRep = (JIluMethodRep) 
        	ilu_GetMethodStubProc(cIluMethod);
        unhand(jjcall)->jNeedsSizing = (Jboolean) ilu_CallNeedsSizing(cCall);
        if (TRACE1) {
            ilu_DebugPrintf("$ IluConnOps_receiveRequest got request\n");
        }
        unhand(jjcall)->needFinish = 1; /* enforces ilu_FinishCall */
        return 1; /*keep looping*/
    } else if (stat == ilu_RcvReqStat_quit) {
        if (TRACE1) {
            ilu_DebugPrintf("$ IluConnOps_receiveRequest quit \n");
        }
        if (initted) ilu_FinishCall(cCall, &lerr);
        ilu_DoneServingConnection(cConn, &lerr);
        return 0; /*exit loop*/
    } else if (stat == ilu_RcvReqStat_noop) {
        if (TRACE1) {
            ilu_DebugPrintf("$ IluConnOps_receiveRequest noop \n");
        }
        if (initted) ilu_FinishCall(cCall, &lerr);
        return 1; /*keep looping*/
    } else {
        if (TRACE1) {
            ilu_DebugPrintf("$ IluConnOps_receiveRequest other \n");
        }
        if (initted) ilu_FinishCall(cCall, &lerr);
        return 1; /*keep looping*/
    }
} /* receiveRequest */


EXPORTLIMITED void
_ilujava_forkConnectionHandler(ilu_Connection newConnection)
{
    Jlong conn;
    if (newConnection) {
        *(ilu_Connection*)&conn = newConnection;
        execute_java_dynamic_method(EE(), 
            (struct Hjava_lang_Object *) jProtoCFInstance,
            "dynamicForkSC",
            "(J)V",
            conn);
    }
}


/* end */

