/* IluJava_IluOInt.c */
/* Chris Jacobi, May 3, 1997 8:54 am PDT */

/*
 * Copyright (c) 1996 Xerox Corporation.  All Rights Reserved.  
 * Unlimited use, reproduction, and distribution of this software is
 * permitted.  Any copy of this software must include both the above
 * copyright notice of Xerox Corporation and this paragraph.  Any
 * distribution of this software must comply with all applicable United
 * States export control laws.  This software is made available AS IS,
 * and XEROX CORPORATION DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED,
 * INCLUDING WITHOUT LIMITATION THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE, AND NOTWITHSTANDING ANY OTHER
 * PROVISION CONTAINED HEREIN, ANY LIABILITY FOR DAMAGES RESULTING FROM
 * THE SOFTWARE OR ITS USE IS EXPRESSLY DISCLAIMED, WHETHER ARISING IN
 * CONTRACT, TORT (INCLUDING NEGLIGENCE) OR STRICT LIABILITY, EVEN IF
 * XEROX CORPORATION IS ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
 */
 
/* 
 * $Id: IluJava_IluOInt.c,v 1.24 1997/05/03 21:20:01 jacobi Exp $ 
 *
 *
 * This class helps in the mapping from real Java objects to kernel object 
 * Native side for IluOInt.java
 */

 
#include "IluJava_JStubs.h"
#include "xerox_ilu_IluOIntContainer.h"

#include "IluJava.h"
#include "IluJava_JTypes.h"
#include <iluhash.h>
#include "IluJava_Ops.h"

#include "IluJava_JMon.h"

static HashTable _cjOb2jOI; 
    /* 
     * key: true and surrogate java ilu object; 
     * val: IluOInt 
     *
     * locked with peekTabLock
     */
     
static ILUJAVA_JMON_PTR _ilujava_peekTabLock; 


static JIluOInt 
_ilujava_internalPeek(JIluObject jjobj) 
{
     JIluOInt jjoi;
     jjoi = (JIluOInt)
     		ilu_hash_FindInTable(_cjOb2jOI, jjobj);
     return jjoi;
}


EXPORTJAVA long
xerox_ilu_IluOInt_setupForPeek(
	JIluOInt jjoi,
	JIluObject jjobj
	) 
    /* 
     * Makes the real object remember the IluOInt.
     * 0 is success.
     * If realObject is already known to Ilu this will
     * return 1 and no state is changed.
     */
{
    if ((jjoi == 0) | (jjobj == 0)) return 1;
    ILUJAVA_MON_ENTER(_ilujava_peekTabLock);
    if (_ilujava_internalPeek(jjobj) != 0) {
        ILUJAVA_MON_EXIT(_ilujava_peekTabLock);
        return 1;
    }
    ilu_hash_AddToTable(_cjOb2jOI, (ilu_refany) jjobj, (ilu_refany) jjoi);
    unhand(jjoi)->jjJavaObject = jjobj;
    ILUJAVA_MON_EXIT(_ilujava_peekTabLock);
    return 0;
}


static void 
internalDisablePeek(JIluOInt jjoi) 
{
     JIluObject jjobj = unhand(jjoi)->jjJavaObject;
     if (jjobj != 0) {
         ilu_hash_RemoveFromTable(_cjOb2jOI, jjobj);
     }
}


EXPORTJAVA JIluOInt 
xerox_ilu_IluOInt_nPeekWithContainer(
	JIluOInt unused,
	JIluObject jjobj,
	struct Hxerox_ilu_IluOIntContainer * jjCont)
/* private procedure; we trust jjCont */ 
{
    JIluOInt jjoi;
    JIluObject jjCheck = 0;
    ILUJAVA_MON_ENTER(_ilujava_peekTabLock);
    jjoi = _ilujava_internalPeek(jjobj);
    if (jjoi) {
        if (unhand(jjoi)->destroyed) {
            jjoi = 0;
        } else {
            jjCheck = unhand(jjoi)->jjJavaObject;
        }
    }
    unhand(jjCont)->jjoi = jjoi;
    ILUJAVA_MON_EXIT(_ilujava_peekTabLock);
    if ((jjoi != 0) & (jjCheck != jjobj)) {
        BUGPRINT("$ IluOInt_nPeekWithContainer HUH?");
        return 0;
    }
    return jjoi;
}


EXPORTJAVA JIluOInt 
xerox_ilu_IluOInt_nPeekWithCall(
	JIluOInt unused,
	JIluObject jjobj,
	JIluCall jCall)
/* 
 * non-public procedure; we trust jCall.
 * Very similar to nPeekWithContainer but uses an IluCall for traceability 
 */ 
{
    JIluOInt jjoi;
    JIluObject jjCheck = 0;
    ILUJAVA_MON_ENTER(_ilujava_peekTabLock);
    jjoi = _ilujava_internalPeek(jjobj);
    if (jjoi) {
        if (unhand(jjoi)->destroyed) {
            jjoi = 0;
        } else {
            jjCheck = unhand(jjoi)->jjJavaObject;
        }
    }
    unhand(jCall)->jDontGCshort = (HObject*) jjoi;
    ILUJAVA_MON_EXIT(_ilujava_peekTabLock);
    if ((jjoi != 0) & (jjCheck != jjobj)) {
        BUGPRINT("$ IluOInt_nPeekWithCall HUH?");
        return 0;
    }
    return jjoi;
}


typedef struct _KInfo {
    ilu_Object cIluObject;
    ilu_Server cServer;
    ilu_Class cClass;
} KInfo;


/* Assigns values to kinfo* and returns kernel object from an IluOInt
 *
 * before: not Inside (kinfo->cServer, kinfo->cClass)
 * after:  return != ILU_NIL => Inside(kinfo->cServer, kinfo->cClass)
 * after:  return == ILU_NIL => not Inside (kinfo->cServer, kinfo->cClass)
 */
static ilu_Object 
getSetCIluObject(JIluOInt jjoi, KInfo* kinfo, ilu_boolean raiseErrs)
{
    JIluClassRep jjClass = unhand(jjoi)->jjClassRep;
    if (jjClass) {
        kinfo->cClass = Access(jjClass, yIluClass, ilu_Class);
    } else {
        kinfo->cClass = ilu_rootClass;
    }
    if (kinfo->cClass == 0) {
        kinfo->cIluObject = 0;
        if (raiseErrs) {
            SignalError(EE(), "xerox/ilu/IluSomeSystemException", 
                "no ilu_Class");
        }
        return 0;
    }
    kinfo->cServer = _ilujava_EnterServer(jjoi, kinfo->cClass);
    if (kinfo->cServer == 0) {
        kinfo->cIluObject = 0;
        if (raiseErrs) {
            SignalError(EE(), "xerox/ilu/IluSomeSystemException", 
                "no ilu_Server");
        }
        return 0;
    }
    kinfo->cIluObject = Access(jjoi, yIluKernelObject, ilu_Object);
    if (kinfo->cIluObject == 0) {
        ilu_ExitServer(kinfo->cServer, kinfo->cClass);
        if (raiseErrs) {
            SignalError(EE(), "xerox/ilu/IluSomeSystemException", 
                "no ilu kernel object");
        }
        return 0;
    }
    return kinfo->cIluObject;
}


EXPORTJAVA long 
xerox_ilu_IluOInt_destroyGentleOI(
	JIluOInt jjoi) 
{
     JIluClassRep jjClass = unhand(jjoi)->jjClassRep;
     ilu_Class cClass = Access(jjClass, yIluClass, ilu_Class);
     ilu_Server cServer = 0;
     ilu_Object cIluObject;
     if (_ilujava_gcFlag > 0) {
         ilu_DebugPrintf("$ destroyGentleOI: looking at %x \n", jjoi);
     }
     if (Access(jjoi, yServer, ilu_Server) == 0) {
         if (_ilujava_gcFlag > 0) {
             ilu_DebugPrintf("$ destroyGentleOI: %x no server 0 \n", jjoi);   
         }
         return 0; /*don't re-finalize*/
     }
     if (unhand(jjoi)->veryInterest OR  
             unhand(jjoi)->retained OR 
             unhand(jjoi)->ghost) {
         /* Not monitored, must be conservative */
         if (_ilujava_gcFlag > 0) {
             ilu_DebugPrintf("$ destroyGentleOI: re-finalize 1 %x \n", jjoi);   
         }
         return 1; /*re-finalize*/
     }
     /* use the server lock */  
     cServer = _ilujava_EnterServer(jjoi, cClass);
     if (cServer == 0) {
         if (_ilujava_gcFlag > 0) {
             ilu_DebugPrintf("$ destroyGentleOI: %x no server 1 \n", jjoi);   
         }
         return 0; /*don't re-finalize*/
     }
     cIluObject = Access(jjoi, yIluKernelObject, ilu_Object);
     if (cIluObject != 0) {
         if (ilu_VeryInterested(cIluObject) OR unhand(jjoi)->ghost) {
             ilu_ExitServer(cServer, cClass);
             if (_ilujava_gcFlag > 0) {
                 ilu_DebugPrintf("$ destroyGentleOI: re-finalize 2 %x \n", jjoi);   
             }
             return 1; /*re-finalize*/
         }
     }
    ILUJAVA_MON_ENTER(_ilujava_peekTabLock);
    if (unhand(jjoi)->veryInterest OR  
            unhand(jjoi)->retained OR  
            unhand(jjoi)->ghost) {
        ILUJAVA_MON_EXIT(_ilujava_peekTabLock);
        ilu_ExitServer(cServer, cClass);
        if (_ilujava_gcFlag > 0) {
             ilu_DebugPrintf("$ destroyGentleOI: re-finalize 4 %x \n", jjoi);   
        }
        return 1; /*re-finalize*/
     }
     ILUJAVA_MON_EXIT(_ilujava_peekTabLock);
     ilu_ExitServer(cServer, cClass);
     /* Exit the locks and re-enter them in destroyRudeOI
      * This is necessary to null-out the yServer field with the
      * global server lock held again.
      */
     xerox_ilu_IluOInt_destroyRudeOI(jjoi);
     return 0; /*don't re-finalize*/
}


EXPORTJAVA void
xerox_ilu_IluOInt_destroyRudeOI(JIluOInt jjoi)
{
     JIluClassRep jjClass = unhand(jjoi)->jjClassRep;
     ilu_Class cClass = Access(jjClass, yIluClass, ilu_Class);
     ilu_Server cServer;
     ilu_Object cIluObject;
     if (_ilujava_gcFlag > 0) {
         ilu_DebugPrintf("$ destroyRudeOI %x \n", jjoi);
     }
     if (unhand(jjoi)->destroyed) {
         if (_ilujava_gcFlag > 2) {
             ilu_DebugPrintf("$ destroyRudeOI already destroyed \n");
         }
         return;
     }
     cServer = _ilujava_EnterServerDisable(jjoi, cClass);
     if (cServer) {
         if (_ilujava_gcFlag > 2) {
             ilu_DebugPrintf("$ destroyRudeOI entered lock %x \n", jjoi);
         }
         ILUJAVA_MON_ENTER(_ilujava_peekTabLock);
         if (_ilujava_gcFlag > 4) {
             ilu_DebugPrintf("$ destroyRudeOI entered peekTabLock %x \n", jjoi);
         }
         cIluObject = Access(jjoi, yIluKernelObject, ilu_Object);
         unhand(jjoi)->destroyed = 1;
         unhand(jjoi)->yIluKernelObject = 0;
         internalDisablePeek(jjoi);
         ILUJAVA_MON_EXIT(_ilujava_peekTabLock);
         if (cIluObject) {
             if (_ilujava_gcFlag > 4) {
                 ilu_DebugPrintf("$ destroyRudeOI before reg 0 \n");
             }
             ilu_RegisterLanguageSpecificObject(cIluObject, 0, JAVALANGIDX);
         }
         if (_ilujava_gcFlag > 4) {
             ilu_DebugPrintf("$ destroyRudeOI before ExitServer");
         }
         ilu_ExitServer(cServer, cClass);
     }
     if (_ilujava_gcFlag > 2) {
         ilu_DebugPrintf("$ destroyRudeOI exit %x \n", jjoi);
     }
}


EXPORTJAVA void
xerox_ilu_IluOInt_withdrawOI(JIluOInt jjoi)
{
    KInfo kinfo;
    char * ckey;
    ilu_boolean huh;
    getSetCIluObject(jjoi, &kinfo, ilu_TRUE);
    if (kinfo.cIluObject) {
         ckey = Access(jjoi, yOwnerKey, char *);
         if (ckey) {
             Access(jjoi, yOwnerKey, char *) = 0;
             huh = ilu_WithdrawObject(kinfo.cIluObject, ckey);
         } else {
             ilu_ExitServer(kinfo.cServer, kinfo.cClass);
         }
    }
}


EXPORTJAVA void
xerox_ilu_IluOInt_publishOI(JIluOInt jjoi)
{
    KInfo kinfo;
    char * ckey;
    getSetCIluObject(jjoi, &kinfo, ilu_TRUE);
    if (kinfo.cIluObject) {
        ckey = Access(jjoi, yOwnerKey, char *);
        if (ckey) {
            ilu_ExitServer(kinfo.cServer, kinfo.cClass);
            SignalError(EE(), "xerox/ilu/IluSomeSystemException", "xxx-204");
            return;
        }
        ckey = ilu_PublishObject(kinfo.cIluObject);
        Access(jjoi, yOwnerKey, char *) = ckey;
    }
}


/* called within server lock */
static ilu_boolean 
_ilujava_noter(ilu_Object cobj, int vi)
{
    JIluOInt jjoi;
    jjoi = (JIluOInt) 
    	ilu_GetLanguageSpecificObject(cobj, JAVALANGIDX);
    if (_ilujava_gcFlag > 0) {
        ilu_DebugPrintf("$ IluOInt noter: %d c<%x> j<%x> \n", vi, cobj, jjoi);
    }
    if (jjoi != ILU_NIL && ilu_TrueInstanceP(cobj)) {
        ilu_Class cl = ilu_ClassOfObject(cobj);
        if (_ilujava_gcFlag>1) {
            ilu_DebugPrintf("$ IluOInt noter is-true\n");
        }
        if ( ilu_CollectibleP(cl)) {
            struct execenv *ee = EE(); /*Java execution environment*/
            if (_ilujava_gcFlag>1) {
                ilu_DebugPrintf("$ IluOInt noter is-collectable\n");
            }
            if (vi) {
                execute_java_dynamic_method(ee,
                    (struct Hjava_lang_Object *) jjoi,
                    "showInterest", 
                    "()V");
            } else {
                execute_java_dynamic_method(ee,
                    (struct Hjava_lang_Object *) jjoi,
                    "removeInterest", 
                    "()V");
            }
            if (ee->exceptionKind !=0) {
                _ilujava_PANIC("$ ObjectAxillary noter: java exception\n");
            }
        }
    }
    return ilu_TRUE;
}				


EXPORTLIMITED void
_ilujava_IluOIntInit()
{
    _ilujava_peekTabLock = ILUJAVA_MON_ALLOC();
    _cjOb2jOI = ilu_hash_MakeNewTable(111, 
    	ilu_hash_HashPointer, 
    	ilu_hash_PointerCompare
    	);
    
    ilu_SetNoter(_ilujava_noter, JAVALANGIDX);
}


EXPORTJAVA Hjava_lang_String *
xerox_ilu_IluOInt_sbhOfOI(JIluOInt jjoi)
{
    KInfo kinfo;
    Hjava_lang_String * jsbh = 0;
    char * csbh;
    getSetCIluObject(jjoi, &kinfo, ilu_TRUE);
    if (kinfo.cIluObject) {
        csbh = ilu_SBHOfObject(kinfo.cIluObject);
        if (csbh) {
            jsbh = makeJavaString(csbh, strlen(csbh));
        }
        ilu_ExitServer(kinfo.cServer, kinfo.cClass);
    }
    return jsbh;
}


EXPORTJAVA Hjava_lang_String *
xerox_ilu_IluOInt_iorOfOI(JIluOInt jjoi)
{
    Hjava_lang_String * jsbh = 0;
#ifdef IIOP_PROTOCOL
    ilu_Error err = ILU_INIT_NO_ERR;
    KInfo kinfo;
    char * csbh;
    getSetCIluObject(jjoi, &kinfo, ilu_TRUE);
    if (kinfo.cIluObject) {
        csbh = ilu_IOROfObject(kinfo.cIluObject, &err);
        if (csbh) {
            jsbh = makeJavaString(csbh, strlen(csbh));
        }
        ilu_ExitServer(kinfo.cServer, kinfo.cClass);
    }
    if (ILU_ERRNOK(err)) {
        _ilujava_IluErrorToCallException(&err, "iorOfOI");
    }
#endif /* IIOP_PROTOCOL */
    return jsbh;
}



EXPORTJAVA void
xerox_ilu_IluOInt_nativePingOI(JIluOInt jjoi)
{
    KInfo kinfo;
    getSetCIluObject(jjoi, &kinfo, ilu_TRUE);
    if (kinfo.cIluObject) {
        ilu_Error err = ILU_INIT_NO_ERR;
        ilu_boolean ok;
        ilu_Connection newConnection;
        err = ilu_DeltaHolds(kinfo.cIluObject, 1); 
                /* as side effect also helds server */
        ilu_ExitServer(kinfo.cServer, kinfo.cClass);
        if (ILU_ERRNOK(err)) {
            _ilujava_IluErrorToCallException(&err, "ping 0");
            return;
        }
        ok = ilu_PingObject(kinfo.cIluObject, &newConnection);
        if (newConnection) {
            /* need to monitor outgoing connection */
            _ilujava_forkConnectionHandler(newConnection);
        }
        ilu_EnterServer(kinfo.cServer, kinfo.cClass);
        err = ilu_DeltaHolds(kinfo.cIluObject, -1);
        ilu_ExitServer(kinfo.cServer, kinfo.cClass);
        if (ILU_ERRNOK(err)) {
            _ilujava_IluErrorToCallException(&err, "ping 1");
        } else if (!ok) {
            _ilujava_IluErrorToCallException(0, "ping 2");
        }
    }
}

/* end */

