/* IluJava_IluServer.c */
/* Chris Jacobi, April 23, 1997 9:12 am PDT */

/*
 * Copyright (c) 1996, 1997 Xerox Corporation.  All Rights Reserved.  
 * Unlimited use, reproduction, and distribution of this software is
 * permitted.  Any copy of this software must include both the above
 * copyright notice of Xerox Corporation and this paragraph.  Any
 * distribution of this software must comply with all applicable United
 * States export control laws.  This software is made available AS IS,
 * and XEROX CORPORATION DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED,
 * INCLUDING WITHOUT LIMITATION THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE, AND NOTWITHSTANDING ANY OTHER
 * PROVISION CONTAINED HEREIN, ANY LIABILITY FOR DAMAGES RESULTING FROM
 * THE SOFTWARE OR ITS USE IS EXPRESSLY DISCLAIMED, WHETHER ARISING IN
 * CONTRACT, TORT (INCLUDING NEGLIGENCE) OR STRICT LIABILITY, EVEN IF
 * XEROX CORPORATION IS ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
 */
 
/* $Id: IluJava_IluServer.c,v 1.11 1997/04/23 17:14:23 jacobi Exp $ */
 
/* 
 * C side for IluServer.java 
 * See also IluObjectTable.java
 */

#include <iluxport.h>
#include <iluerror.h>
/* for debugging only */
#include <iluntrnl.h>

#include "IluJava_JStubs.h"
#include "IluJava.h"
#include "IluJava_JMon.h"

/* The Global server lock protects existence of ilu_server
 * before an ilu_server is used for locking operations..
 * Enter JGServerlock before entering server locks.
 */
static ILUJAVA_JMON_PTR gServerLock = 0;
#define ENTER_GSERVER_LOCK() ILUJAVA_MON_ENTER(gServerLock)
#define EXIT_GSERVER_LOCK() ILUJAVA_MON_EXIT(gServerLock)


EXPORTLIMITED void _ilujava_IluServerInit()
{
    gServerLock = ILUJAVA_MON_ALLOC();
}


EXPORTLIMITED ilu_Server
_ilujava_EnterServer(JIluOInt jjoi, ilu_Class cActualClass)
/* return 0: no lock entered
 * return ilu_Server: server lock entered
 */
{
    ilu_Server cServer;
    ENTER_GSERVER_LOCK();
    cServer = Access(jjoi, yServer, ilu_Server); 
    if (cServer) {
        ilu_EnterServer(cServer, cActualClass);
    }
    EXIT_GSERVER_LOCK();
    return cServer;
}


EXPORTLIMITED ilu_Server
_ilujava_EnterServer2(
    struct Hxerox_ilu_IluServer * jjIluServer, ilu_Class cActualClass)
/* return 0: no lock entered
 * return ilu_Server: server lock entered
 */
{
    ilu_Server cServer;
    ENTER_GSERVER_LOCK();
    cServer = Access(jjIluServer, yIluServer, ilu_Server); 
    if (cServer) {
        ilu_EnterServer(cServer, cActualClass);
    }
    EXIT_GSERVER_LOCK();
    return cServer;
}


EXPORTLIMITED ilu_Server
_ilujava_EnterServerDisable(JIluOInt jjoi, ilu_Class cActualClass)
/* return 0: no lock entered
 * return ilu_Server: server lock entered
 */
{
    ilu_Server cServer;
    ENTER_GSERVER_LOCK();
    cServer = Access(jjoi, yServer, ilu_Server); 
    if (cServer) {
        ilu_EnterServer(cServer, cActualClass);
        Access(jjoi, yServer, ilu_Server) = 0;
    }
    EXIT_GSERVER_LOCK();
    return cServer;
}


/* enters and exits the server lock */
EXPORTLIMITED ilu_Server
_ilujava_ServerEnterHolds(JIluOInt jjoi)
/* increases holds on server with appropriate locks.
 * if returns 0: no holds changed
 * if returns ilu_Server: server enter+exit and holds changed
 */
{
    ILU_ERRS((bad_locks, internal/broken)) err;
    ilu_Server cServer;
    ENTER_GSERVER_LOCK();
    cServer = Access(jjoi, yServer, ilu_Server); 
    if (cServer) {
        ilu_EnterServer(cServer, ilu_rootClass);
        ilu_DeltaServerHolds(cServer, 1, &err);
        ilu_ExitServer(cServer, ilu_rootClass);
    }
    EXIT_GSERVER_LOCK();
    return cServer;
}


/* enters and exits the server lock */
EXPORTLIMITED void
_ilujava_ServerExitHolds(ilu_Server cServer)
/* decreases holds on server with appropriate locks.
 */
{
    ILU_ERRS((bad_locks, internal/broken)) err;
    if (cServer) {
        ilu_EnterServer(cServer, ilu_rootClass);
        ilu_DeltaServerHolds(cServer, -1, &err);
        ilu_ExitServer(cServer, ilu_rootClass);
    }
}


/* L1 >= {server}; L2, Main unconstrained */
INTERNALONLY ilu_Object
_ilujava_ObjectTableCreateObj(ilu_ObjectTable cot, ilu_string cih)
/* Upcalled by kernel to request creation of an object */
{
    ilu_Server cServer;
    ilu_Class cClass;
    ilu_Object cObject;
    struct Hxerox_ilu_IluClassRep * jjIluClass;
    struct Hxerox_ilu_IluObjectTable * jjObjTab;
    struct Hxerox_ilu_IluServer * jjServer;
    struct Hxerox_ilu_IluOInt * jjoi;
    struct Hjava_lang_String * jjih;
    jjObjTab = (struct Hxerox_ilu_IluObjectTable *) cot->ot_rock;
    jjServer = unhand(jjObjTab)->jjServer;
    jjih = /*javaString.h*/ makeJavaString((char *) cih, strlen(cih));
    if (_ilujava_objectsFlag>1) {
        ilu_DebugPrintf("$ ObjectTableCreateObj for %s jjServer %x\n", cih, jjServer);
    }
    jjoi = (struct Hxerox_ilu_IluOInt *)
			execute_java_dynamic_method(EE(),
        (struct Hjava_lang_Object *) jjObjTab,
        "doCreateTrueObject", 
        "(Ljava/lang/String;)Lxerox/ilu/IluOInt;",
        jjih
        );
    if (_ilujava_objectsFlag>1) {
        ilu_DebugPrintf("$ ObjectTableCreateObj jjoi=<%x>\n", jjoi);
    }
    /* conceptionally similar (locking would be wrong) to 
     * xerox_ilu_IluRT0_registerTrue(0, jjoi, jjih, jjServer);
     */
    if (jjoi) {
        jjIluClass = unhand(jjoi)->jjClassRep;
        cClass = Access(jjIluClass, yIluClass, ilu_Class);
        cServer = Access(jjServer, yIluServer, ilu_Server); /* hold lock */
        Access(jjoi, yServer, ilu_Server) = cServer; /* hold lock */
        if (_ilujava_objectsFlag>1) {
            ilu_DebugPrintf("$ ObjectTableCreateObj before FindOrCreateTrueObject cServer=<%x>\n", cServer);
        }
        cObject = ilu_FindOrCreateTrueObject(cih, cServer, cClass, jjoi);
        Access(jjoi, yIluKernelObject, ilu_Object) = cObject;
        if (_ilujava_objectsFlag>1) {
            ilu_DebugPrintf("$ ObjectTableCreateObj cObject=<%x>\n", cObject);
        }
        return cObject;
    }
    _ilujava_notUsed1 = (char *) jjih;
    return 0;
}


INTERNALONLY void
_ilujava_ObjectTableFreeSelf(ilu_ObjectTable cot)
/* Upcalled by kernel to request deletion of table */
{
    execute_java_dynamic_method(EE(),
        (struct Hjava_lang_Object *) cot->ot_rock,
        "doObjectTableFreed", 
        "()V"
        );
    ilu_free(cot);
}


INTERNALONLY ilu_ObjectTable _ilujava_createOT(
	struct Hxerox_ilu_IluObjectTable * jjObjTab
	)
/* Create the native side of an object table */
{
    ilu_ObjectTable cot;
    if (jjObjTab==0) return 0;
    cot = ilu_must_malloc(sizeof(ilu_ObjectTable_s));
    cot->ot_object_of_ih = _ilujava_ObjectTableCreateObj;
    cot->ot_free_self = _ilujava_ObjectTableFreeSelf;
    cot->ot_rock = jjObjTab; /*not GC'ed because the server references it*/
    return cot;
}


EXPORTJAVA void 
xerox_ilu_IluServer_nativeOfCreateServer(
	struct Hxerox_ilu_IluServer * jjIluServer, 
	struct Hjava_lang_String * jjID
	)    
{
    ILU_ERRS((bad_locks, broken_locks, bad_param, no_memory, internal)) err;
    ilu_ObjectTable cObjtab = 0; 
    ilu_Server cIluServer;
    char * cId = /* javaString.h [malloc] */ allocCString(jjID); 
    struct Hxerox_ilu_IluObjectTable * jjObjTab = unhand(jjIluServer)->jjObjTab;
    if (jjObjTab) {
        cObjtab = _ilujava_createOT(jjObjTab);
    }
    cIluServer = ilu_CreateTrueServer(cId, cObjtab, JAVALANGIDX, &err);
    if (cIluServer==0) BUGPRINT("$ Couldn't create true server");
    ILU_MUST_BE_SUCCESS(err);
    Access(jjIluServer, yIluServer, ilu_Server) = cIluServer;
    Access(jjIluServer, yPtr, char *) = cId;
    (void) ilu_SetLSS(cIluServer, jjIluServer, JAVALANGIDX, &err);
    ILU_MUST_BE_SUCCESS(err);
    ilu_ExitServer(cIluServer, ilu_rootClass);
    if (_ilujava_pctFlag>0) {
        ilu_DebugPrintf("$ IluServer_nativeOfCreateServer: true server created; j<%x> c<%x>  \n", jjIluServer, cIluServer);
    }
} /*nativeOfCreateServer*/


EXPORTJAVA void 
xerox_ilu_IluServer_nativeOfDestroyServer(
    struct Hxerox_ilu_IluServer * jjIluServer)
{
    ILU_ERRS((bad_param, internal)) err;
    ilu_Server cIluServer;
    if (_ilujava_pctFlag>0) {
        ilu_DebugPrintf("$ IluServer_nativeOfDestroyServer; j<%x>\n", jjIluServer);
    }
    ENTER_GSERVER_LOCK();
    cIluServer = Access(jjIluServer, yIluServer, ilu_Server);
    Access(jjIluServer, yIluServer, ilu_Server) = 0;
    if (cIluServer) {
        ilu_EnterServer(cIluServer, ilu_rootClass);
    }
    EXIT_GSERVER_LOCK();
    if (cIluServer) {
        ilu_InnerBankServer(cIluServer);
        ilu_SetLSS(cIluServer, 0, JAVALANGIDX, &err);
        ILU_MUST_BE_SUCCESS(err);
        ilu_ExitServer(cIluServer, ilu_rootClass);
    }
} /*nativeOfDestroyServer*/


EXPORTJAVA void 
xerox_ilu_IluServer_nativeOfFreeServer(
    struct Hxerox_ilu_IluServer * jjIluServer)
{
    char * cId = Access(jjIluServer, yPtr, char *);
    Access(jjIluServer, yPtr, char *) = 0;
    ilu_free(cId);
} /*nativeOfFreeServer*/


/* end */
